package com.javacodegeeks.generate_zip_file;

import java.io.IOException;
import java.nio.channels.Channels;
import java.nio.channels.FileChannel;
import java.nio.channels.WritableByteChannel;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.nio.file.StandardOpenOption;
import java.util.Objects;
import java.util.concurrent.ExecutorService;
import java.util.zip.ZipEntry;
import java.util.zip.ZipOutputStream;

import javax.servlet.AsyncContext;
import javax.servlet.ServletException;
import javax.servlet.annotation.WebInitParam;
import javax.servlet.annotation.WebServlet;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

@WebServlet(urlPatterns = "/zipfile", initParams = { @WebInitParam(name = "src", value = "/tmp/sample.mp4"), @WebInitParam(name = "timeout", value = "5000") }, asyncSupported = true)
public final class ZipFileServlet extends HttpServlet {

    private static final long serialVersionUID = 1L;
    private static final String TIMEOUT_PARAM_KEY = "timeout";
    private static final String SRC_PATH = "src";

    @Override
    protected void doGet(final HttpServletRequest request, final HttpServletResponse response) throws ServletException, IOException {
        final AsyncContext asyncCtx = request.startAsync();
        asyncCtx.setTimeout(Long.valueOf(getServletConfig().getInitParameter(TIMEOUT_PARAM_KEY)));

        getExecutorService(request).execute(new ZipFileHandler(response, getServletConfig().getInitParameter(SRC_PATH)));
    }

    private ExecutorService getExecutorService(final HttpServletRequest request) {
        assert !Objects.isNull(request);
        return (ExecutorService) request.getServletContext().getAttribute(ContextListenerExecutorServiceInitializer.THREAD_POOL_EXECUTOR);
    }

    private static final class ZipFileHandler implements Runnable {

        private static final long MAX_BYTE_COUNT = 67076096l;
        private static final String ZIP_CONTENT_TYPE = "application/zip";
        private static final String CONTENT_DISPOSITION_KEY = "Content-Disposition";
        private static final String CONTENT_DISPOSITION_VALUE = "attachment; filename=\"%s\"";

        private final HttpServletResponse response;
        private final Path srcPath;

        private ZipFileHandler(final HttpServletResponse response, final String srcPath) {
            assert !Objects.isNull(response) && !Objects.isNull(srcPath);

            this.response = response;
            this.srcPath = Paths.get(srcPath);
        }

        @Override
        public void run() {
            try (ZipOutputStream zipStream = new ZipOutputStream(this.response.getOutputStream());
                    WritableByteChannel outputChannel = Channels.newChannel(zipStream);
                    FileChannel inputChannel = FileChannel.open(this.srcPath, StandardOpenOption.READ);) {

                final String fileName = this.srcPath.getFileName().toString();

                initResponse(response, fileName);

                zipStream.putNextEntry(new ZipEntry(fileName));

                transferContents(inputChannel, outputChannel);

                zipStream.flush();
                this.response.getOutputStream().flush();
            } catch (IOException e) {
                e.printStackTrace();
            } finally {
                try {
                    this.response.getOutputStream().close();
                } catch (IOException e) {
                    e.printStackTrace();
                }
            }
        }

        private void initResponse(final HttpServletResponse response, final String fileName) {
            assert !Objects.isNull(response) && !Objects.isNull(fileName) && !fileName.isEmpty();

            this.response.setContentType(ZIP_CONTENT_TYPE);
            this.response.setHeader(CONTENT_DISPOSITION_KEY, String.format(CONTENT_DISPOSITION_VALUE, fileName + ".zip"));
        }

        private void transferContents(final FileChannel inputChannel, final WritableByteChannel outputChannel) throws IOException {
            long size = inputChannel.size();
            long position = 0;
            while (position < size) {
                position += inputChannel.transferTo(position, MAX_BYTE_COUNT, outputChannel);
            }
        }
    }
}
